package hstar.creole.parser

import collection.mutable.ListBuffer
import hstar.creole.syntaxtree.{VoidElement, ForkElement, Element}

/**
 * Developed with pleasure :)<br>
 * User: HoD<br>
 * Date: 02.10.2010<br>
 * Time: 19:50:02<br>
 */

class IntermediateSyntaxNode(val syntax: SyntaxRule = SyntaxRules.defaultSyntax, beginsAt: Int = 0, val parent: Option[IntermediateSyntaxNode] = None) {
  private var currentSubString    = ""
  private val children            = new ListBuffer[IntermediateSyntaxNode]
  private var parsedElements      = Option.empty[Seq[Element]]
  private var parsedSubStrings    = new ListBuffer[String]
  private var consumedBeginMarker = Option.empty[String]
  private var consumedEndMarker   = Option.empty[String]

  def newNode(syntax: SyntaxRule, atIndex: Int) = {
    val child = new IntermediateSyntaxNode(syntax, atIndex, Some(this))
    children += child
    child
  }

  def mostRecentlyParsedSubString = {
    parsedSubStrings.lastOption
  }

  def setBeginMarker(marker: String) {
    assert(consumedBeginMarker.isEmpty)
    consumedBeginMarker = Some(marker)
  }

  def setEndMarker(marker: String) {
    assert(consumedEndMarker.isEmpty)
    consumedEndMarker = Some(marker)
  }

  def getBeginMarker = consumedBeginMarker.get

  def getEndMarker = consumedEndMarker.get

  def addChar(c: Char) {
    currentSubString += c
  }

  def addCut() {
    parsedSubStrings += currentSubString
    currentSubString = ""
  }

  def createSyntaxElement() {
    assert(parsedElements.isEmpty)
    parsedElements = Some(parsedSubStrings.map(e => syntax.toElement(e, this)))
  }

  def toElement: Element = {
    val elements = {
      val childElements = children.view.filterNot(_.isInstanceOf[VoidElement]).map(_.toElement).toList
      var foldedAndDechunked = syntax.foldParsedAndChildElements(parsedElements.getOrElse(Nil), childElements)
      if (foldedAndDechunked.size > 1) {
        FoldingRules.allLiveRules.foreach(e => {
          foldedAndDechunked = e.foldElements(foldedAndDechunked)
        })
      }
      foldedAndDechunked
    }
    if (elements.size == 1) {
      elements.head
    } else {
      new ForkElement(elements)
    }
  }
 
  protected def applyToThisAndParents[U](f: (IntermediateSyntaxNode) => U) {
    f(this)
    parent.foreach(_.applyToThisAndParents(f))
  }

  def selfAndAllParents: Traversable[IntermediateSyntaxNode] = {
    new Traversable[IntermediateSyntaxNode] {
      def foreach[U](f: (IntermediateSyntaxNode) => U) = {
        applyToThisAndParents(f)
      }
    }
  }
  
  def allParents: Traversable[IntermediateSyntaxNode] = selfAndAllParents.view.drop(1)

  def mostRecentlyAttachedNode(ignoreChildren: Boolean) = {
    if (ignoreChildren) {
      children.lastOption
    } else {
      children.lastOption.map(_.traverseForNewestNode)
    }
  }

  private def traverseForNewestNode: IntermediateSyntaxNode = {
    if (children.isEmpty) {
      this
    } else {
      children.last.traverseForNewestNode
    }
  }

  override def toString = {
    "markers: " + consumedBeginMarker.getOrElse("") + ", " + consumedEndMarker.getOrElse("") + ", elements: " + parsedElements + ", " + syntax
  }
}


