package hstar.creole.parser

import annotation.elidable
import annotation.elidable._
import java.util.logging.{Level, Logger}

/**
 * Developed with pleasure :)<br>
 * User: HoD<br>
 * Date: 02.10.2010<br>
 * Time: 19:47:02<br>
 */

class Parser private(completeSource: String) {
  
  val syntaxTree = {
    val logger = Logger.getAnonymousLogger
    
    @elidable(3000)
    def log(s: => String) {
     // logger.log(Level.INFO, s)
      print(s)
    }
    val cursor = new Cursor(completeSource.length - 1)
    val incompleteTree = new IntermediateSyntaxNode
    def parse(ctx: ParsingContext) {
      @elidable(3000)
      def logln(s: String) {
        // logger.log(Level.INFO, s)
        println(s + ", current cursor is at " + cursor.currentIndex + ", text to parse is '" + ctx.remainingText+"'")
      }
      //are chars left to parse on the highest level?
      while (cursor.hasNext) {
        //find the matching syntax
        val syntax = SyntaxRules.allRules.find(_.canBegin(ctx)).getOrElse(SyntaxRules.defaultSyntax)
        logln("\nsyntax is "+syntax)
        //create a new node for the current element
        val newNode = ctx.currentIncompleteElement.newNode(syntax, cursor.currentIndex)
        //parse char by char until the current node is done
        var currentSyntaxStillApplies = true
        //now leep going until the current syntax no longer applies (-> return to upper level) or the parsing is finished (while loop ends)
        var markerConsumed = false
        //we need to know if this is the first iteration after returning from a lower level 
        var justReturnedFromLowerLevel = false
        var addNewNodeNow = false
        
        def endCurrentElementParsing() {
          logln("\nfinished parsing an element")
          if (cursor.hasNext) {
            val exitMarker = syntax.skipSyntaxExitMarker(cursor, ctx)
            logln("exit marker of "+syntax+" is '"+exitMarker+"'")
            newNode.setEndMarker(exitMarker)
          }
          newNode.addCut()
          newNode.createSyntaxElement()
          logln("newest parsed element is "+newNode.mostRecentlyParsedSubString)
        }
        def incCursorIndex() {
          if (!markerConsumed) {
            //consume the marker (but only once)
            val beforeConsumption = cursor.currentIndex
            val beginMarker = syntax.skipSyntaxInitMarker(cursor, ctx)
            newNode.setBeginMarker(beginMarker)
            markerConsumed = true
            logln("consumed marker '"+beginMarker+"'")
          } else {
            if (!justReturnedFromLowerLevel) {
              cursor.incCursorIndex()
            } else {
              logln("skipping cursor inc")
              justReturnedFromLowerLevel = false
            }
          }
        }
        def handleLowerLevel() = {
          //tell the node that the current level either ends here or gets interrupted by a nested element
          newNode.addCut()
          parse(new ParsingContext(newNode, cursor.view, completeSource))
          //set the "do not increase the cursor index one time"-flag. the lower level just stopped at a char that could not be parsed by the lower level syntax, but might be 
          //by the current one
          justReturnedFromLowerLevel = true
          //upon return, check if the current rule still applies and/or should skip some characters in case its endAllBranches-Condition was true
          currentSyntaxStillApplies = syntax.canContinue(ctx)
          //check the current syntax if it caused all branches to end (not the whole tree!).
          syntax.endsAllBranchesAt(ctx)    
        }
        def parseLowerLevelAndContinue() {
          logln("switching to lower level")
          val endedByEndBranchRule = handleLowerLevel()
          //there are 2 possible reasons for the return from the lower level
          if (endedByEndBranchRule) {
            //the current syntax is no longer valid and just cut all branches. finish the parsing here and open a new node at the same level.
            logln("parsing ended by branch ending rule of "+syntax)
            endCurrentElementParsing()
            addNewNodeNow = true
          } else {
            //otherwise, we're finished and there's nothing left to parse. we'll just let the control flow get us out of this method :)
            logln("branch ending rule of " + syntax+" was false")
          }
          logln("returned from lower level back to "+syntax)
        }
        def continueAtSameLevel() {
          logln("\nfinished current element, staying at the same level")
          newNode.addCut()
          newNode.createSyntaxElement()
          addNewNodeNow = true
        }
        do {
          if (currentSyntaxStillApplies) {
            incCursorIndex()
            val endByBranchRule = ctx.allParents.exists(_.syntax.endsAllBranchesAt(ctx)) 
            if (endByBranchRule) {
              //some parent says we should stop here and return to the upper level
              currentSyntaxStillApplies = false
            } else {
              //no reason to stop
              val nestedRuleMatch = SyntaxRules.allRulesAndDefault.find(e => {
                val syntaxDifferent = e != syntax
                def canBeNested = syntax.canDirectlyNestElementOfType(e)
                def subSyntaxCanBegin = e.canBegin(ctx)
                syntaxDifferent && canBeNested && subSyntaxCanBegin
              })
              val successorRuleMatch = SyntaxRules.allRulesAndDefault.find(e => {
                val syntaxDifferent = e != syntax
                def switchToTypePossible = syntax.canSwitchToElementOfType(e)
                def subSyntaxCanbegin = e.canBegin(ctx)
                syntaxDifferent && switchToTypePossible && subSyntaxCanbegin
              })
              assert(nestedRuleMatch.isEmpty || successorRuleMatch.isEmpty)
              currentSyntaxStillApplies = syntax.canContinue(ctx)
              if (!currentSyntaxStillApplies) {
                logln("\nsyntax no longer applies: " + syntax)
              }
              if (currentSyntaxStillApplies && nestedRuleMatch.isEmpty && successorRuleMatch.isEmpty) {
                //if the current syntax is the only valid one: add the current char
                log("+" + completeSource.charAt(cursor.currentIndex))
                newNode.addChar(completeSource.charAt(cursor.currentIndex))
              } else {
                //depending on which one of these options is defined, the next iteration will continue after the subelement is parsed
                nestedRuleMatch.foreach(_ => parseLowerLevelAndContinue())
                //or right now
                successorRuleMatch.foreach(_ => continueAtSameLevel())
                //if none of the above is defined, we're finished here and the control flow will direct us out of the method
              }
            }
          } else {
            //the current syntax is no longer valid. finish the parsing here and return to the upper level
            endCurrentElementParsing()
            return
          }
        }
        while (cursor.hasNext && !addNewNodeNow)
        if (!addNewNodeNow) {
          //after parsing all valid characters on the highest level, create a syntax element out of them
          endCurrentElementParsing()
        }
      }
    }
    parse(new ParsingContext(incompleteTree, cursor.view, completeSource))
    incompleteTree.createSyntaxElement()
    incompleteTree.toElement
  }
}

object Parser {
  def apply(source: String) = new Parser(source).syntaxTree
} 