package hstar.creole.syntaxtree

/**
 * Developed with pleasure :)<br>
 * User: HoD<br>
 * Date: 25.09.2010<br>
 * Time: 22:15:59<br>
 */

abstract class Element {
  def children: Seq[Element]
}

class VoidElement(desc:String = "dummy element, to be removed") extends Element {
  def children = Nil
}

class PlaceholderElement[T](desc:String = "placeholder", val info:Option[T] = None) extends Element {
  def children = Nil
  
  override def toString = "<placeholder>"+desc+"<placeholder>"
}

class WrappingElement(wrapped: Element) extends Element {
  override def toString = "WrappingElement(" + wrapped + ")"

  def children = List(wrapped)
}

case object ForkElement {
  def apply(left: Element, right: Element):ForkElement = {
    ForkElement(List(left, right))
  }
}

case class ForkElement(subElements: Seq[Element]) extends Element {
  def this(left:Element, right:Element) {
    this(List(left, right))
  }
  
  def asUnforkedIfOnlyOneElement = {
    if (subElements.size==1) {
      subElements.head
    } else {
      this
    }
  }
  
  override def toString = "ForkElement(" + subElements + ")"

  def children = subElements
}

case class SimpleText(text: String) extends Element {
  override def toString = "SimpleText(\"" + text + "\")"

  def children = Nil
}

case class ForcedLineBreak extends Element {
  override def toString = "LineBreak()"

  def children = Nil
}

case class SimpleLineBreak extends Element {
  override def toString = "SimpleLineBreak()"

  def children = Nil
}

case class Paragraph extends Element {
  override def toString = "Paragraph()"

  def children = Nil
}

case class LinkAndDescription(text: Option[String], url: String) extends Element {
  override def toString = "LinkAndDescription(Some(\"" + text + "\", \"" + url + "\")"

  def children = Nil
}

case class AttributedElement(wrapped: Element, attribute: Attribute) extends WrappingElement(wrapped) {
  override def toString = {
    "AttributedElement(" + wrapped + ", " + attribute + ")"
  }
}

case class Table(wrapped:Element) extends WrappingElement(wrapped) {
  override def toString = "Table("+wrapped+")"
}

case class TableRow(wrapped:Element) extends WrappingElement(wrapped) {
  override def toString = "TableRow("+wrapped+")"
}

case class TableCell(wrapped:Element) extends WrappingElement(wrapped) {
  override def toString = "TableCell("+wrapped+")"
}

case class TableHeaderCell(wrapped:Element) extends WrappingElement(wrapped) {
  override def toString = "TableHeader("+wrapped+")"
}

object AttributedElement {
  def apply(s: String, attribute: Attribute) = {
    new AttributedElement(new SimpleText(s), attribute)
  }
}